from uuid import uuid4

import pandas as pd

from psdmpy.models.enums import ElectricCurrentType
from psdmpy.models.input.create.utils import create_data
from psdmpy.models.input.participant import evcs
from psdmpy.models.input.participant.em import EnergyManagementSystems
from psdmpy.models.input.participant.evcs import EvChargingStations
from psdmpy.models.input.participant.evs import ElectricVehicles
from psdmpy.models.input.participant.hp import HeatPumps
from psdmpy.models.input.participant.load import Loads
from psdmpy.models.input.participant.pv import PhotovoltaicPowerPlants
from psdmpy.models.input.participant.storage import Storages


def fixed_q_characteristics(cos_phi):
    return f"cosPhiFixed:{{(0.0, {cos_phi})}}"


def create_pvs(data_dict):
    return PhotovoltaicPowerPlants(create_data(data_dict, create_pv_data))


def create_pv_data(
    id,
    node,
    s_rated,
    azimuth,  # South = 0°, West = 90°, East = -90°
    elevation_angle,  # Tilted inclination from horizontal [0°, 90°]
    eta_conv=97,
    cos_phi_rated=0.9,
    albedo=0.21,
    q_characteristics=None,
    k_g=0.9,
    k_t=1,
    market_reaction=False,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(cos_phi_rated)
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "node": node,
            "q_characteristics": q_characteristics,
            "s_rated": s_rated,
            "albedo": albedo,
            "azimuth": azimuth,
            "elevation_angle": elevation_angle,
            "eta_conv": eta_conv,
            "k_g": k_g,
            "k_t": k_t,
            "market_reaction": market_reaction,
            "cos_phi_rated": cos_phi_rated,
        }
    ).rename(uuid)


def create_storages(data_dict):
    return Storages(create_data(data_dict, create_storages_data))


def create_storages_data(
    id,
    node,
    e_storage,
    s_rated=None,
    q_characteristics=None,
    behaviour="self",
    type_uuid=None,
    type_id=None,
    capex=0,
    opex=0,
    cos_phi_rated=0.9,
    p_max=None,
    active_power_gradient=100,
    eta=100,  # efficiency of the electrical inverter
    dod=90,  # depth of discharge -> dod% of the capacity can be used
    life_time=9999999,  # permissible hours of full use
    life_cycle=9999999,  # permissible amount of full cycles
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not s_rated:
        s_rated = e_storage / 2
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(cos_phi_rated)
    if not type_uuid:
        type_uuid = str(uuid4())
    if not type_id:
        type_id = f"Type_{id}"
    if not p_max:
        p_max = s_rated
    if not uuid:
        uuid = str(uuid4())
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "node": node,
            "q_characteristics": q_characteristics,
            "type_uuid": type_uuid,
            "type_id": type_id,
            "capex": capex,
            "opex": opex,
            "s_rated": s_rated,
            "cos_phi_rated": cos_phi_rated,
            "p_max": p_max,
            "active_power_gradient": active_power_gradient,
            "e_storage": e_storage,
            "eta": eta,
            "dod": dod,
            "life_time": life_time,
            "life_cycle": life_cycle,
        }
    ).rename(uuid)


def create_electric_vehicles(data_dict):
    return ElectricVehicles(create_data(data_dict, create_electric_vehicles_data))


def create_electric_vehicles_data(
    id,
    node,
    s_rated,
    e_storage,
    e_cons,
    q_characteristics=None,
    type_uuid=None,
    type_id=None,
    capex=0,
    opex=0,
    cos_phi_rated=1,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    if not type_uuid:
        type_uuid = str(uuid4())
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(cos_phi_rated)
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "node": node,
            "q_characteristics": q_characteristics,
            "type_uuid": type_uuid,
            "type_id": type_id,
            "capex": capex,
            "opex": opex,
            "s_rated": s_rated,
            "cos_phi_rated": cos_phi_rated,
            "e_storage": e_storage,
            "e_cons": e_cons,
        }
    ).rename(uuid)


def create_ev_charging_stations(data_dict):
    return EvChargingStations(create_data(data_dict, create_ev_charging_stations_data))


def create_ev_charging_stations_data(
    id,
    node,
    location_type,
    s_rated,
    v2g_support=False,
    charging_points=1,
    electric_current_type=ElectricCurrentType.AC,
    cos_phi=0.9,
    type_id=None,
    q_characteristics=None,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(cos_phi)
    if not type_id:
        type_id = id
    type = evcs.evcs_type.substitute(
        id=id, s_rated=s_rated, current_type=electric_current_type.value
    )
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "node": node,
            "q_characteristics": q_characteristics,
            "cos_phi_rated": cos_phi,
            "charging_points": charging_points,
            "location_type": location_type,
            "type": type,
            "v2g_support": v2g_support,
        }
    ).rename(uuid)


def create_heat_pumps_data(
    id,
    node,
    thermal_bus,
    s_rated,
    p_thermal=None,
    q_characteristics=None,
    type_uuid=None,
    type_id=None,
    capex=0,
    opex=0,
    cos_phi_rated=0.9,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not p_thermal:
        p_thermal = s_rated * 3
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(cos_phi_rated)
    if not type_uuid:
        type_uuid = str(uuid4())
    if not type_id:
        type_id = f"Type_{id}"
    if not uuid:
        uuid = str(uuid4())
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "node": node,
            "q_characteristics": q_characteristics,
            "thermal_bus": thermal_bus,
            "type_uuid": type_uuid,
            "type_id": type_id,
            "capex": capex,
            "opex": opex,
            "s_rated": s_rated,
            "cos_phi_rated": cos_phi_rated,
            "p_thermal": p_thermal,
        }
    ).rename(uuid)


def create_heat_pumps(data_dict):
    return HeatPumps(create_data(data_dict, create_heat_pumps_data))


def create_loads_data(
    id,
    node,
    s_rated,
    e_cons_annual,
    load_profile,
    dsm=True,
    q_characteristics=None,
    cos_phi_rated=0.9,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(cos_phi_rated)
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "node": node,
            "q_characteristics": q_characteristics,
            "load_profile": load_profile,
            "dsm": dsm,
            "e_cons_annual": e_cons_annual,
            "s_rated": s_rated,
            "cos_phi_rated": cos_phi_rated,
        }
    ).rename(uuid)


def create_loads(data_dict):
    return Loads(create_data(data_dict, create_loads_data))


def create_energy_management_systems_data(
    id,
    node,
    connected_assets,
    control_strategy,
    q_characteristics=None,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    if not q_characteristics:
        q_characteristics = fixed_q_characteristics(0.9)
    return pd.Series(
        {
            "id": id,
            "node": node,
            "connected_assets": connected_assets,
            "control_strategy": control_strategy,
            "q_characteristics": q_characteristics,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
        }
    ).rename(uuid)


def create_energy_management_systems(data_dict):
    return EnergyManagementSystems(
        create_data(data_dict, create_energy_management_systems_data)
    )



from dataclasses import dataclass
from typing import List

from pandas import Series

from psdmpy.models.enums import RawGridElementsEnum
from psdmpy.models.input.connector.connector import Connector
from psdmpy.models.input.container.mixins import HasTypeMixin


@dataclass(frozen=True)
class Lines(HasTypeMixin, Connector):
    @property
    def length(self) -> Series:
        return self.data["length"]

    @property
    def geo_position(self) -> Series:
        return self.data["geo_position"]

    @property
    def olm_characteristic(self) -> Series:
        return self.data["olm_characteristic"]

    @property
    def v_rated(self) -> Series:
        return self.data["v_rated"]

    @property
    def r(self) -> Series:
        return self.data["r"]

    @property
    def x(self) -> Series:
        return self.data["x"]

    @property
    def b(self) -> Series:
        return self.data["b"]

    @property
    def i_max(self) -> Series:
        return self.data["i_max"]

    def aggregated_line_length(self) -> float:
        """
        Returns the aggregated length of all lines.

        Returns:
            float: Aggregated length of all lines.
        """
        return self.data["length"].sum()

    def relative_line_length(self) -> float:
        """
        Returns the relative length of all lines.

        Returns:
            float: Relative length of all lines.
        """
        return self.data["length"] / len(self.data)

    @staticmethod
    def get_enum() -> RawGridElementsEnum:
        return RawGridElementsEnum.LINE

    @staticmethod
    def entity_attributes() -> List[str]:
        return [
            "length",
            "geo_position",
            "olm_characteristic",
        ]

    @staticmethod
    def type_attributes() -> List[str]:
        return HasTypeMixin.type_attributes() + ["r", "x", "b", "g", "i_max", "v_rated"]



import os
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from pandas import DataFrame
from pandas.core.groupby import DataFrameGroupBy

ROOT_DIR = os.path.abspath(__file__ + "/../../../")


class DateTimePattern(Enum):
    UTC_TIME_PATTERN_EXTENDED = "%Y-%m-%dT%H:%M:%SZ"
    UTC_TIME_PATTERN = "%Y-%m-%dT%H:%MZ"
    PLAIN = "%Y-%m-%d %H:%M:%S"


def get_absolute_path_from_project_root(path: str):
    if not isinstance(path, str):
        path = str(path)
    if path.startswith(ROOT_DIR):
        return path
    else:
        return Path(ROOT_DIR).joinpath(path)


def get_absolute_path_from_working_dir(path: str | Path) -> Path:
    """
    Given a path (as string or pathlib.Path), returns its absolute path based on
    the current working directory. If the path is already absolute, it's returned unchanged.

    Args:
    - path (Union[str, Path]): The input path.

    Returns:
    - Path: The absolute path as a pathlib.Path object.
    """
    path_obj = Path(path)
    if path_obj.is_absolute():
        return path_obj
    return path_obj.resolve()


def get_file_path(path: str | Path, file_name: str):
    return Path(path).resolve().joinpath(file_name)


def read_csv(
    path: str | Path, file_name: str, delimiter: str, index_col: Optional[str] = None
) -> DataFrame:
    full_path = get_file_path(path, file_name)
    if not full_path.exists():
        raise IOError("File with path: " + str(full_path) + " does not exist")
    if index_col:
        return pd.read_csv(
            full_path, delimiter=delimiter, quotechar='"', index_col=index_col
        )
    else:
        return pd.read_csv(full_path, delimiter=delimiter, quotechar='"')


def to_date_time(zoned_date_time: str) -> datetime:
    """
    Converts zoned date time string with format: "yyyy-MM-dd'T'HH:mm:ss[.S[S][S]]'Z'"
    e.g. '2022-02-01T00:15Z[UTC]' to python datetime

    Args:
        zoned_date_time: The zoned date time string to convert.

    Returns:
        The converted datetime object.
    """
    if not zoned_date_time or not isinstance(zoned_date_time, str):
        raise ValueError(f"Unexpected date time string: {zoned_date_time}")
    try:
        year = int(zoned_date_time[0:4])
        month = int(zoned_date_time[5:7])
        day = int(zoned_date_time[8:10])
        hour = int(zoned_date_time[11:13])
        minute = int(zoned_date_time[14:16])
    except IndexError:
        raise IOError(f"Could not parse time stamp: {zoned_date_time}")
    return datetime(year=year, month=month, day=day, hour=hour, minute=minute)


def csv_to_grpd_df(
    file_name: str, simulation_data_path: str, delimiter: str
) -> DataFrameGroupBy:
    """
    Reads in a PSDM csv results file cleans it up and groups it by input_archive model.

    Args:
        file_name: name of the file to read
        simulation_data_path: base directory of the result data
        delimiter: the csv delimiter

    Returns:
        DataFrameGroupBy object of the file
    """
    data = read_csv(simulation_data_path, file_name, delimiter)

    if "uuid" in data.columns:
        data = data.drop(columns=["uuid"])
    return data.groupby(by="input_model")


def check_filter(filter_start: Optional[datetime], filter_end: Optional[datetime]):
    if (filter_start or filter_end) and not (filter_start and filter_end):
        raise ValueError(
            "Both start and end of the filter must be provided if one is provided."
        )
    if (filter_start and filter_end) and (filter_start > filter_end):
        raise ValueError("Filter start must be before end.")


def df_to_csv(
    df: DataFrame,
    path: Union[str, Path],
    file_name: str,
    mkdirs=True,
    delimiter: str = ",",
    index_label="uuid",
    datetime_pattern=DateTimePattern.UTC_TIME_PATTERN,
):
    df = df.copy(deep=True)
    if isinstance(path, Path):
        path = str(path)
    file_path = get_file_path(path, file_name)
    if mkdirs:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

    bool_cols = []
    for col in df.columns:
        is_bool_col = df[col].dropna().map(type).eq(bool).all()
        if is_bool_col:
            bool_cols.append(col)

    # replace True with 'true' only in boolean columns
    df[bool_cols] = df[bool_cols].replace({True: "true", False: "false"})

    if isinstance(df.index, pd.DatetimeIndex):
        df.index = df.index.strftime(datetime_pattern.value)

    datetime_cols = df.select_dtypes(
        include=["datetime64[ns, UTC]", "datetime64"]
    ).columns
    for col in datetime_cols:
        df[col] = df[col].apply(
            lambda x: x.strftime(datetime_pattern.value) if not pd.isnull(x) else x
        )

    df.to_csv(file_path, index=True, index_label=index_label, sep=delimiter)


def bool_converter(maybe_bool):
    if isinstance(maybe_bool, bool):
        return maybe_bool
    elif isinstance(maybe_bool, str) and maybe_bool.lower() in ["true", "false"]:
        return maybe_bool.lower() == "true"
    else:
        raise ValueError("Cannot convert to bool: " + str(maybe_bool))



from dataclasses import dataclass
from typing import List

from pandas import Series

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.result.entity import ResultEntities
from psdmpy.models.result.power import PQResult
from psdmpy.processing.series import add_series, hourly_mean_resample


@dataclass(frozen=True)
class FlexOptionResult(ResultEntities):
    def __add__(self, other: "FlexOptionResult"):
        p_ref_sum = add_series(self.p_ref(), other.p_ref(), "p_ref")
        p_min_sum = add_series(self.p_min(), other.p_min(), "p_min")
        p_max_sum = add_series(self.p_max(), other.p_max(), "p_max")
        summed_data = p_ref_sum.to_frame().join([p_min_sum, p_max_sum])
        return FlexOptionResult(
            self.entity_type,
            "FlexResult - Sum",
            "FlexResult - Sum",
            summed_data,
        )

    @staticmethod
    def attributes() -> List[str]:
        return ["p_max", "p_min", "p_ref"]

    def p_max(self):
        return self.data["p_max"]

    def p_min(self):
        return self.data["p_min"]

    def p_ref(self):
        return self.data["p_ref"]

    def p_max_as_pq(self, sp_type: SystemParticipantsEnum) -> PQResult:
        return self._p_to_pq_res(sp_type, self.p_max())

    def p_ref_as_pq(self, sp_type: SystemParticipantsEnum) -> PQResult:
        return self._p_to_pq_res(sp_type, self.p_ref())

    def p_min_as_pq(self, sp_type: SystemParticipantsEnum):
        return self._p_to_pq_res(sp_type, self.p_min())

    def _p_to_pq_res(
        self, sp_type: SystemParticipantsEnum, p_series: Series
    ) -> PQResult:
        data = p_series.rename("p").to_frame()
        data["q"] = 0
        return PQResult(sp_type, "flex-signal", "flex-signal", data)

    def hourly_resample(self):
        updated_data = self.data.apply(lambda x: hourly_mean_resample(x))
        return FlexOptionResult(
            self.entity_type, self.name, self.input_model, updated_data
        )

    def add_series(self, series: Series) -> "FlexOptionResult":
        updated_data = self.data.apply(
            lambda p_flex: add_series(p_flex, series, p_flex.name), axis=0
        )
        return FlexOptionResult(
            self.entity_type, self.name, self.input_model, updated_data
        )



import numbers
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Type

import numpy as np
import pandas as pd
from pandas import DataFrame, Series

from psdmpy.models.enums import RawGridElementsEnum
from psdmpy.models.input.entity import Entities
from psdmpy.models.input.node import Nodes
from psdmpy.models.result.entity import ResultEntities
from psdmpy.models.result.participant.dict import ResultDict, ResultDictType


@dataclass(frozen=True)
class NodeResult(ResultEntities):
    def __eq__(self, other):
        if not isinstance(other, NodeResult):
            return False
        return (
            (self.input_model == other.input_model)
            & (self.name == other.name)
            & (self.data.equals(other.data))
        )

    @property
    def v_mag(self) -> Series:  # in Ampere
        return self.data["v_mag"]

    @property
    def v_ang(self) -> Series:
        return self.data["v_ang"]

    def v_complex(self, v_rated_kv_src: numbers.Real | Nodes) -> Series:
        v_rated_kv = (
            v_rated_kv_src
            if isinstance(v_rated_kv_src, numbers.Real)
            else v_rated_kv_src.subset(self.input_model).v_rated.iloc[0]
        )
        return (self.v_mag * v_rated_kv) * np.exp(1j * np.radians(self.v_ang))

    @staticmethod
    def attributes() -> List[str]:
        return ["v_ang", "v_mag"]

    # todo: fix me
    @staticmethod
    def build_from_nominal_data(
        uuid: str,
        name: Optional[str],
        data: DataFrame,
        rated_voltage: float,
    ) -> "NodeResult":
        data["v_mag"] = data["v_mag"].divide(rated_voltage)
        return NodeResult(RawGridElementsEnum.NODE, uuid, name, data)


@dataclass(frozen=True)
class NodesResult(ResultDict):
    entities: dict[str, NodeResult]

    @classmethod
    def from_csv(
        cls: Type[ResultDictType],
        simulation_data_path: str,
        delimiter: str,
        simulation_end: Optional[datetime] = None,
        input_entities: Optional[Entities] = None,
        filter_start: Optional[datetime] = None,
        filter_end: Optional[datetime] = None,
    ) -> ResultDictType:
        return super().from_csv(
            RawGridElementsEnum.NODE,
            simulation_data_path,
            delimiter,
            simulation_end,
            input_entities,
            filter_start,
            filter_end,
        )

    @property
    def v_mag(self) -> Optional[DataFrame]:
        if not self.entities:
            return None
        return pd.concat(
            [
                node_res.v_mag.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def v_ang(self) -> Optional[DataFrame]:
        if not self.entities:
            return None
        return pd.concat(
            [
                node_res.v_ang.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    def v_mag_describe(self) -> DataFrame:
        return self._describe(self.v_mag)

    def v_ang_describe(self) -> DataFrame:
        return self._describe(self.v_ang)

    @staticmethod
    def _describe(data: DataFrame):
        return data.describe().transpose()

    @classmethod
    def create_empty(cls, entity_type=RawGridElementsEnum.NODE) -> "NodesResult":
        return super().create_empty(entity_type)



from functools import partial

from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.result.container.participants import (
    ParticipantsResultContainer,
)
from psdmpy.models.result.flex_option import FlexOptionResult
from psdmpy.models.result.power import PQResult
from psdmpy.plots.common.line_plot import ax_plot_time_series
from psdmpy.plots.common.utils import (
    FIGSIZE,
    FLEX_MAX,
    FLEX_MIN,
    FLEX_REF,
    ORANGE,
    set_title,
)
from psdmpy.plots.results.power_plot import ax_plot_active_power


def plot_all_participants_flex_range(
    participants_res_container: ParticipantsResultContainer,
    title: str,
    resolution: str,
    hourly_mean: bool = False,
    include_actual_res=False,
):
    em_res_list = [
        res
        for res in participants_res_container.to_list(
            include_em=False, include_flex=False, include_empty=False
        )
        if res.entity_type != SystemParticipantsEnum.LOAD
    ]

    plot_count = len(em_res_list)
    fig, axs = plt.subplots(plot_count, 1, figsize=(10, 11), sharex=True, sharey=False)

    # plt.tight_layout()
    plt.subplots_adjust(hspace=0.5)

    for idx, participant_res in enumerate(em_res_list):
        participant_flex = participants_res_container.flex.subset(
            participant_res.entities.keys()
        )
        if participant_flex:
            flex_sum = participant_flex.sum()
            axs[idx].set_title(f"{participant_res.entity_type.get_plot_name()}")
            ax_plot_flex_range(axs[idx], flex_sum, "d", hourly_mean=hourly_mean)
        if include_actual_res:
            ax_plot_active_power(
                axs[idx],
                participant_res.sum(),
                resolution,
                hourly_mean=hourly_mean,
                color=ORANGE,
                label="p_actual",
            )
        axs[idx].legend()
        axs[idx].set_ylabel("Power in MW")

    return fig, axs


def plot_flex_range(
    flex_option: FlexOptionResult,
    title: str,
    resolution: str,
    hourly_mean: bool,
    actual_res: PQResult = None,
):
    figure, ax = plt.subplots(figsize=FIGSIZE)
    # plt.tight_layout()
    ax_plot_flex_range(ax, flex_option, resolution, hourly_mean, actual_res)
    title = title if title else "Flex Options"

    set_title(ax, title)
    return figure, ax


def ax_plot_flex_range(
    ax: Axes,
    flex_option: FlexOptionResult,
    resolution: str,
    hourly_mean: bool,
    actual_res: PQResult = None,
):
    p_ref = flex_option.p_ref()
    plot_func = partial(
        ax_plot_time_series,
        ax=ax,
        type=SystemParticipantsEnum.FLEX_OPTIONS,
        resolution=resolution,
        hourly_mean=hourly_mean,
    )

    plot_func(
        res=flex_option.p_max(),
        label="p_max",
        color=FLEX_MAX,
        fill_between=p_ref,
    )
    plot_func(
        res=flex_option.p_min(),
        label="p_min",
        color=FLEX_MIN,
        fill_between=p_ref,
    )
    plot_func(res=p_ref, label="p_ref", color=FLEX_REF)

    if actual_res:
        ax_plot_active_power(
            ax,
            actual_res,
            resolution,
            hourly_mean=hourly_mean,
            color=ORANGE,
            label="p_actual",
        )

    ax.set_ylabel("Power in MW")

    ax.legend()



from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Optional, Union

from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.input.container.participants import (
    SystemParticipantsContainer,
)
from psdmpy.models.input.container.raw_grid import RawGridContainer
from psdmpy.models.primary_data import PrimaryData
from psdmpy.models.result.power import PQResult


@dataclass(frozen=True)
class GridContainer(ContainerMixin):
    raw_grid: RawGridContainer
    # TODO: we keep the participant containers effectively twice with the mapping
    participants: SystemParticipantsContainer
    primary_data: PrimaryData
    node_participants_map: Dict[str, SystemParticipantsContainer]

    def __bool__(self):
        return len(self.to_list(include_primary_data=True)) > 0

    def to_list(self, include_empty: bool = False, include_primary_data: bool = False):
        grid = (
            [self.raw_grid, self.participants, self.primary_data]
            if include_primary_data
            else [self.raw_grid, self.participants]
        )
        return grid if include_empty else [g for g in grid if g]

    def get_nodal_primary_data(self):
        time_series = []
        nodal_primary_data = dict()
        for node, participants_container in self.node_participants_map.items():
            participants_uuids = participants_container.uuids().tolist()
            node_primary_data = self.primary_data.get_for_participants(
                participants_uuids
            )
            time_series.extend(node_primary_data)
            node_primary_data_agg = PQResult.sum(node_primary_data)
            nodal_primary_data[node] = node_primary_data_agg
        return nodal_primary_data

    def get_nodal_sp_count_and_power(self):
        data = {}
        for node_uuid, sps in self.node_participants_map.items():
            nodal_data = {}
            for sp in sps.to_list(include_empty=False):
                sp_id = sp.get_enum().value
                count = len(sp.data)
                data_str = f"Count: {count}"
                # check if sp has a property named s_rated
                if hasattr(sp, "s_rated"):
                    s_rated = round(sp.s_rated.sum(), 2)
                    data_str += f", Rated Power: {s_rated} kw"
                nodal_data[sp_id] = data_str
            data[node_uuid] = nodal_data
        return data

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        return GridContainer(
            raw_grid=self.raw_grid,
            participants=self.participants,
            primary_data=self.primary_data.filter_by_date_time(time),
            node_participants_map=self.node_participants_map,
        )

    def to_csv(
        self,
        path: str,
        include_primary_data: bool,
        mkdirs: bool = True,
        delimiter: str = ",",
    ):
        for container in self.to_list(include_primary_data=include_primary_data):
            try:
                container.to_csv(path, mkdirs=mkdirs, delimiter=delimiter)
            except Exception as e:
                raise IOError(f"Could not write {container} to {path}.", e)

    @classmethod
    def from_csv(
        cls, path: str, delimiter: str, primary_data_delimiter: Optional[str] = None
    ):
        if not primary_data_delimiter:
            primary_data_delimiter = delimiter
        raw_grid = RawGridContainer.from_csv(path, delimiter)
        participants = SystemParticipantsContainer.from_csv(path, delimiter)
        node_participants_map = participants.build_node_participants_map(raw_grid.nodes)
        primary_data = PrimaryData.from_csv(path, primary_data_delimiter)
        return cls(raw_grid, participants, primary_data, node_participants_map)

    @classmethod
    def create_empty(cls):
        return cls(
            raw_grid=RawGridContainer.create_empty(),
            participants=SystemParticipantsContainer.create_empty(),
            primary_data=PrimaryData.create_empty(),
            node_participants_map=dict(),
        )



class ComparisonError(Exception):
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.differences = errors if errors else []
        self.message = message

    def __str__(self):
        error = f"{super().__str__()}\n"
        differences = "\n\n".join(str(d) for d in self.differences)
        return error + differences



import seaborn as sns
from matplotlib import pyplot as plt


def set_sns_style(style: str = "whitegrid", context: str = "notebook"):
    sns.set_style(style)
    sns.set_context(context)


def set_gruvbox_minor_dark_hard():
    dark_tick_color = "#838077"
    tick_color = "#EEEBE1"
    background_color = "#1d2021"
    custom_params = {
        "figure.facecolor": background_color,
        "figure.dpi": 300,
        "figure.figsize": [12, 4],
        "axes.facecolor": background_color,
        "axes.edgecolor": background_color,
        "axes.labelcolor": tick_color,
        "axes.titlecolor": tick_color,
        "xtick.color": tick_color,
        "ytick.color": tick_color,
        "axes.grid": True,
        "grid.color": dark_tick_color,
        "grid.linestyle": "--",
        "grid.linewidth": 0.5,
        "legend.facecolor": background_color,
        "legend.edgecolor": tick_color,
        "legend.labelcolor": tick_color,
        "text.color": tick_color,
    }

    plt.rcParams.update(custom_params)



from dataclasses import dataclass

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class EnergyManagementSystems(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.ENERGY_MANAGEMENT

    @property
    def connected_assets(self):
        return self.data["connected_assets"]

    @property
    def control_strategy(self):
        return self.data["control_strategy"]

    def uuid_to_connected_assets(self):
        return self.connected_assets.to_dict()

    def to_csv(self, path: str, mkdirs=True, delimiter: str = ","):
        self.data["connected_assets"] = self.connected_assets.apply(
            lambda x: f"{' '.join(x)}"
        )
        return super().to_csv(path, mkdirs, delimiter)

    @staticmethod
    def attributes():
        return SystemParticipants.attributes() + [
            "connected_assets",
            "control_strategy",
        ]



import os
from dataclasses import dataclass
from typing import List

from pandas import DataFrame

from psdmpy.io.utils import df_to_csv
from psdmpy.models.enums import RawGridElementsEnum
from psdmpy.models.input.entity import Entities


@dataclass(frozen=True)
class Nodes(Entities):
    data: DataFrame

    @property
    def geo_position(self):
        return self.data["geo_position"]

    @property
    def longitude(self):
        return self.data["longitude"]

    @property
    def latitude(self):
        return self.data["latitude"]

    @property
    def slack(self):
        return self.data["slack"]

    @property
    def subnet(self):
        return self.data["subnet"]

    @property
    def v_rated(self):
        return self.data["v_rated"]

    @property
    def v_target(self):  # in kV
        return self.data["v_target"]

    @property
    def volt_lvl(self):
        return self.data["volt_lvl"]

    @property
    def node(self):
        return self.data.index

    def get_slack_nodes(self):
        return Nodes(self.data[self.slack])

    def to_csv(self, path: str, mkdirs=True, delimiter: str = ","):
        # filter columns latitude and longitude as they are not part of the original data model
        data = self.data.drop(columns=["latitude", "longitude"])
        if mkdirs:
            os.makedirs(os.path.normpath(path), exist_ok=True)
        df_to_csv(data, path, self.get_enum().get_csv_input_file_name(), delimiter)

    @staticmethod
    def get_enum() -> RawGridElementsEnum:
        return RawGridElementsEnum.NODE

    @classmethod
    def attributes(cls) -> List[str]:
        return Entities.attributes() + [
            "v_rated",
            "v_target",
            "slack",
            "geo_position",
            "volt_lvl",
            "subnet",
        ]



from dataclasses import dataclass
from enum import Enum
from string import Template

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class EvChargingStations(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.EV_CHARGING_STATION

    @property
    def charging_points(self):
        return self.data["charging_points"]

    @property
    def location_type(self):
        return self.data["location_type"]

    @property
    def electric_current_type(self):
        return self.data["type"]

    @property
    def v2g_support(self):
        return self.data["v2g_support"]

    @property
    def cos_phi_rated(self):
        return self.data["cos_phi_rated"]

    def get_public_evcs(self):
        return self.data[
            self.location_type.isin(
                [EvcsLocationType.CUSTOMER_PARKING.value, EvcsLocationType.WORK.value]
            )
        ]

    def get_home_evcs(self):
        return self.data[self.location_type.isin([EvcsLocationType.HOME.value])]

    @staticmethod
    def attributes() -> list[str]:
        return SystemParticipants.attributes() + [
            "charging_points",
            "location_type",
            "type",
            "v2g_support",
            "cos_phi_rated",
        ]


class EvcsLocationType(Enum):
    HOME = "HOME"
    WORK = "WORK"
    CUSTOMER_PARKING = "CUSTOMER_PARKING"
    STREET = "STREET"
    CHARGING_HUB_TOWN = "CHARGING_HUB_TOWN"
    CHARGING_HUB_HIGHWAY = "CHARGING_HUB_HIGHWAY"


evcs_type = Template("$id($s_rated|$current_type)")



import logging

from matplotlib.axes import Axes
from pandas import Series

from psdmpy.models.enums import EntitiesEnum
from psdmpy.plots.common.utils import (
    FILL_ALPHA,
    add_to_kwargs_if_not_exist,
    get_label_and_color_dict,
    plot_resample,
    set_date_format_and_label,
)


def ax_plot_time_series(
    ax: Axes,
    res: Series,
    type: EntitiesEnum,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index=False,
    fill_between=None,
    set_x_label: bool = True,
    **kwargs,
):
    args = get_label_and_color_dict(type)
    kwargs = add_to_kwargs_if_not_exist(kwargs, args)
    if set_x_label:
        set_date_format_and_label(ax, resolution)
    try:
        res = plot_resample(res, hourly_mean)
    except TypeError as e:
        logging.warn(
            f"Could not resample time series. Plotting without resampling. Error: {e}"
        )
    ax.plot(res, **kwargs)
    if fill_from_index:
        ax.fill_between(res.index, res, alpha=FILL_ALPHA, color=kwargs["color"])
    elif fill_between is not None:
        ax.fill_between(
            res.index,
            plot_resample(fill_between, hourly_mean),
            res,
            alpha=FILL_ALPHA,
            color=kwargs["color"],
        )
    return ax



import os.path
import uuid
from dataclasses import dataclass
from typing import Dict, Optional

import pandas as pd
from pandas import DataFrame, Series

from psdmpy.errors import ComparisonError
from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.participant.participant import (
    SystemParticipantsWithCapacity,
)
from psdmpy.models.result.participant.dict import ResultDict
from psdmpy.models.result.power import PQResult, PQWithSocResult
from psdmpy.processing.dataframe import join_dataframes


@dataclass(frozen=True)
class PQResultDict(ResultDict):
    entities: Dict[str, PQResult]

    @property
    def p(self) -> Optional[DataFrame]:
        if not self.entities.values():
            return None
        return (
            pd.DataFrame({p_uuid: res.p for p_uuid, res in self.entities.items()})
            .fillna(method="ffill")
            .sort_index()
        )

    @property
    def q(self) -> Optional[DataFrame]:
        if not self.entities.values():
            return None
        return (
            pd.DataFrame({p_uuid: res.q for p_uuid, res in self.entities.items()})
            .fillna(method="ffill")
            .sort_index()
        )

    def p_sum(self) -> Series:
        if not self.entities:
            return Series(dtype=float)
        return self.p.fillna(method="ffill").sum(axis=1).rename("p_sum")

    def q_sum(self):
        if not self.entities:
            return Series(dtype=float)
        return self.q.fillna(method="ffill").sum(axis=1).rename("q_sum")

    def sum(self) -> PQResult:
        return PQResult.sum(list(self.entities.values()))

    def energy(self):
        # todo make concurrent
        sum = 0
        for participant in self.entities.values():
            sum += participant.energy()
        return sum

    def get(self, key: str) -> PQResult:
        return self.entities[key]

    def subset(self, uuids):
        return type(self)(
            self.entity_type,
            {uuid: self.entities[uuid] for uuid in self.entities.keys() & uuids},
        )

    def load_and_generation(self):
        return self.sum().load_and_generation()

    def compare(self, other):
        errors = []

        # Compare time series
        self_keys = set(self.entities.keys())
        other_keys = set(other.entities.keys())
        entity_differences = self_keys.symmetric_difference(other_keys)
        if entity_differences:
            errors.append(
                ComparisonError(
                    f"Differences in entity keys. Following keys not in both dicts: {entity_differences}"
                )
            )

        key_intersection = self_keys & other_keys
        for key in key_intersection:
            try:
                self.entities[key].compare(other.entities[key])
            except ComparisonError as e:
                errors.append(e)

        if errors:
            raise ComparisonError(
                f"Found Differences in {type(self)} comparison: ", errors=errors
            )

    def to_csv(self, path: str, resample_rate: Optional[str] = None):
        file_name = self.entity_type.get_csv_result_file_name()

        def resample(data: DataFrame, input_model: str):
            data = (
                data.resample("60s").ffill().resample(resample_rate).mean()
                if resample_rate
                else data
            )
            data["uuid"] = data.apply(lambda _: uuid.uuid4(), axis=1)
            data["input_model"] = input_model
            return data

        resampled = [
            resample(participant.data, input_model)
            for input_model, participant in self.entities.items()
        ]
        df = join_dataframes(resampled)
        df.to_csv(os.path.join(path, file_name))


@dataclass(frozen=True)
class PQWithSocResultDict(PQResultDict):
    entity_type: SystemParticipantsEnum
    entities: Dict[str, PQWithSocResult]

    def sum_with_soc(self, inputs: SystemParticipantsWithCapacity) -> PQWithSocResult:
        if not self.entities:
            return PQWithSocResult.create_empty(self.entity_type, "", "")
        capacity_participant = []
        for participant_uuid, res in self.entities.items():
            capacity = inputs.get(participant_uuid)[inputs.capacity_attribute()]
            capacity_participant.append((capacity, res))
        return PQWithSocResult.sum_with_soc(capacity_participant)



from dataclasses import dataclass

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.container.mixins import SpTypeMixin
from psdmpy.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class WindEnergyConverters(SpTypeMixin, SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.WIND_ENERGY_CONVERTER

    @property
    def market_reaction(self):
        return self.data["market_reaction"]

    @property
    def eta_conv(self):
        return self.data["eta_conv"]

    @property
    def rotor_area(self):
        return self.data["rotor_area"]

    @property
    def hub_height(self):
        return self.data["hub_height"]

    @staticmethod
    def entity_attributes() -> list[str]:
        return ["market_reaction"]

    @staticmethod
    def type_attributes() -> list[str]:
        return SpTypeMixin.type_attributes() + [
            "cp_characteristic",
            "eta_conv",
            "rotor_area",
            "hub_height",
        ]
